package org.randombits.utils.collection;

import org.apache.commons.collections.IteratorUtils;
import org.randombits.utils.lang.API;

import java.util.*;

/**
 * Provides some utility methods for testing and converting objects that are
 * iterable.
 */
@API
public final class CollectionUtils {

    private CollectionUtils() {
    }


    /**
     * Returns <code>true</code> if the value is either an {@link java.util.Iterator},
     * {@link Iterable}, {@link java.util.Enumeration} or an array.
     *
     * @param value The value to check.
     * @return <code>true</code> if the value is iterable.
     */
    @API
    public static boolean isIterable( Object value ) {
        return value instanceof Iterator || value instanceof Iterable || value instanceof Enumeration
                || value != null && value.getClass().isArray();
    }

    /**
     * Returns the value as an {@link Iterator}. If the value is itself an
     * iterator, an instance of {@link Iterable}, or is an array, the items in
     * the value will be returned as the list. Otherwise, if the value is not
     * <code>null</code>, it will be returned as the only item in the
     * iterator. If the value is <code>null</code>, an empty iterator is
     * returned.
     *
     * @param value The value to convert.
     * @return The iterator.
     */
    @API
    public static Iterator<?> asIterator( Object value ) {
        if ( value instanceof Iterator ) {
            return (Iterator<?>) value;
        } else if ( value instanceof Iterable<?> ) {
            return ( (Iterable<?>) value ).iterator();
        } else if ( value instanceof Enumeration ) {
            return IteratorUtils.asIterator( (Enumeration<?>) value );
        } else if ( value != null ) {
            if ( value.getClass().isArray() ) {
                return IteratorUtils.arrayIterator( (Object[]) value );
            } else {
                return Arrays.asList( value ).iterator();
            }
        } else {
            return IteratorUtils.emptyIterator();
        }
    }

    /**
     * Returns the value as a {@link java.util.Collection}, which may or may not be
     * mutable. It's safest to assume you will not be able to modify the
     * results. The result will never be <code>null</code>.
     *
     * @param value The value to convert to a collection.
     * @return The collection.
     */
    @API
    public static Collection<?> asCollection( Object value ) {
        if ( value instanceof Collection ) {
            return (Collection<?>) value;
        } else if ( value instanceof Iterable ) {
            return IteratorUtils.toList( ( (Iterable<?>) value ).iterator() );
        } else if ( value instanceof Iterator ) {
            return IteratorUtils.toList( (Iterator<?>) value );
        } else if ( value instanceof Enumeration ) {
            return Collections.list( (Enumeration<?>) value );
        } else if ( value != null ) {
            if ( value.getClass().isArray() )
                return Arrays.asList( (Object[]) value );
            else
                return Arrays.asList( value );
        } else {
            return Collections.EMPTY_LIST;
        }
    }

}
