package org.randombits.utils.date;

import org.joda.time.*;
import org.randombits.utils.lang.API;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class provides utility methods for dealing with Dates, in particular with Joda-Time Dates.
 */
@API
public final class DateUtils {
    private enum PeriodType {
        YEARS( DurationFieldType.years(), DateTimeFieldType.year(), 0 ),
        MONTHS( DurationFieldType.months(), DateTimeFieldType.monthOfYear(), 1 ),
        WEEKS( DurationFieldType.weeks(), DateTimeFieldType.weekOfWeekyear(), 1 ),
        DAYS( DurationFieldType.days(), DateTimeFieldType.dayOfMonth(), 1 ),
        HOURS( DurationFieldType.hours(), DateTimeFieldType.hourOfDay(), 0 ),
        MINUTES( DurationFieldType.minutes(), DateTimeFieldType.minuteOfDay(), 0 ),
        SECONDS( DurationFieldType.seconds(), DateTimeFieldType.secondOfMinute(), 0 );

        private final DurationFieldType durationFieldType;
        private final DateTimeFieldType dateTimeFieldType;
        private final int baseValue;

        PeriodType( DurationFieldType durationFieldType, DateTimeFieldType dateTimeFieldType, int baseValue ) {
            this.durationFieldType = durationFieldType;
            this.dateTimeFieldType = dateTimeFieldType;
            this.baseValue = baseValue;
        }

        public DurationFieldType getDurationFieldType() {
            return durationFieldType;
        }

        public static PeriodType forKey( String key, PeriodType lastPeriod ) {
            List<PeriodType> matches = new ArrayList<PeriodType>();
            key = key.toUpperCase();
            for ( PeriodType type : values() ) {
                if ( type.name().startsWith( key ) )
                    matches.add( type );
            }

            if ( matches.size() == 1 )
                return matches.get( 0 );
            else if ( matches.size() == 0 )
                return null;

            // Figure out which one to pick based on 'lastPeriod'.
            if ( lastPeriod != null ) {
                for ( PeriodType type : matches ) {
                    // Return the first type which is after the last period
                    if ( type.compareTo( lastPeriod ) > 0 )
                        return type;
                }
            }
            return matches.get( 0 );
        }

        public DateTimeFieldType getDateTimeFieldType() {
            return dateTimeFieldType;
        }

        public int getBaseValue() {
            return baseValue;
        }
    }

    private enum Shift {
        NEAREST( "" ), UP( "+" ), DOWN( "-" );

        private final String key;

        Shift( String key ) {
            this.key = key;
        }

        public static Shift find( String key ) {
            for ( Shift shift : values() ) {
                if ( shift.key.equals( key ) )
                    return shift;
            }
            return null;
        }
    }

    private static final Pattern PERIOD_PATTERN = Pattern.compile( "\\s*([\\-\\+]?)((?:\\s*[0-9]+\\s*[a-zA-Z]+)+)\\s*" );

    private static final Pattern PERIOD_SECTION_PATTERN = Pattern.compile( "\\s*([0-9]+)\\s*([a-zA-Z]+)" );

    private static final Pattern ROUND_PATTERN = Pattern.compile( "\\s*([\\-\\+]?)\\s*([0-9]+)\\s*([a-zA-Z]+)\\s*" );

    /**
     * Returns <code>true</code> if the provided string matches a valid 'period pattern'.
     * See {@link #getPeriod(String)} for details.
     *
     * @param periodString The period string to test.
     * @return <code>true</code> if it is valid.
     */
    @API
    public static boolean isPeriod( String periodString ) {
        return PERIOD_PATTERN.matcher( periodString ).matches();
    }

    /**
     * Converts a string value into a {@link Period} which can be used to manipulate date values.
     * It is a combination of one or more "+/-X" values, where 'X' can be:
     * <p/>
     * <ul>
     * <li><b>D</b> - Day</li>
     * <li><b>M</b> - Month</li>
     * <li><b>Y</b> - Year</li>
     * <li><b>W</b> - Week</li>
     * <li><b>H</b> - Hour</li>
     * <li><b>S</b> - Second</li>
     * </ul>
     * <p/>
     * The letters are case-insensitive. These can be prefixed by either "+" (the default) or "-", and more than one can be used in a single period string. For example:
     * <p/>
     * <table>
     * <tr><th>String</th><th>Period</th><th>Original</th><th>Result</th></tr>
     * <tr><td>+1D</td><td>Add 1 day</td><td>01-Jan-2010</td><td>02-Jan-2010</td></tr>
     * <tr><td>-5y</td><td>Subtract 5 years</td><td>01-Jan-2010</td><td>01-Jan-2005</td></tr>
     * <tr><td>+1y 6m</td><td>Add 1 year and 6 months</td><td>01-Jan-2010</td><td>01-Jul-2011</td></tr>
     * <tr><td>-1y 1d</td><td>Subtract 1 year and 1 day</td><td>01-Jan-2010</td><td>31-Dec-2008</td></tr>
     * </table>
     *
     * @param periodString The string value.
     * @return The {@link Period} value.
     * @throws DatePeriodException If there is a problem while formatting the string.
     */
    @API
    public static Period getPeriod( String periodString ) throws DatePeriodException {
        MutablePeriod period = Period.ZERO.toMutablePeriod();

        Matcher periodMatcher = PERIOD_PATTERN.matcher( periodString );
        if ( !periodMatcher.matches() )
            throw new DatePeriodException( "Invalid period: " + periodString );

        Shift shift = Shift.find( periodMatcher.group( 1 ) );
        int modifier = Shift.DOWN.equals( shift ) ? -1 : 1;

        Matcher matcher = PERIOD_SECTION_PATTERN.matcher( periodMatcher.group( 2 ) );
        PeriodType type = null;
        while ( matcher.find() ) {
            int amount = parseAmount( matcher.group( 1 ) );
            type = PeriodType.forKey( matcher.group( 2 ), type );
            if ( type == null )
                throw new DatePeriodException( "Invalid period type: " + matcher.group( 2 ) );
            period.add( type.getDurationFieldType(), amount * modifier );
        }

        return period.toPeriod();
    }

    /**
     * Returns a new {@link DateTime} instance that has been shifted by the
     * specified period string. See {@link #getPeriod(String)} for details.
     *
     * @param dateTime     The date.
     * @param periodString The period to shift by.
     * @return The shifted date.
     * @throws DatePeriodException if there is an issue while formatting the string.
     */
    @API
    public static ReadableDateTime shiftDate( ReadableDateTime dateTime, String periodString ) {
        Period period = getPeriod( periodString );
        return dateTime.toDateTime().plus( period );
    }

    /**
     * Returns a new {@link java.util.Date} instance that has been shifted by the
     * specified period string. See {@link #getPeriod(String)} for details.
     *
     * @param date   The date.
     * @param period The period to shift by.
     * @return The shifted date.
     * @throws DatePeriodException if there is an issue while formatting the string.
     */
    @API
    public static Date shiftDate( Date date, String period ) {
        return shiftDate( new DateTime( date ), period ).toDateTime().toDate();
    }

    /**
     * Rounds the provided date to the nearest value for the provided period string.
     *
     * <p>Examples</p>
     *
     * <table>
     *     <tr>
     *         <td>Date</td>
     *         <th>Period</th>
     *         <th>Return Value</th>
     *     </tr>
     *     <tr>
     *         <td>1-Jan-2011 01:07:11</td>
     *         <td>"15min"</td>
     *         <td>1-Jan-2011 01:00:00</td>
     *     </tr>
     *     <tr>
     *         <td>1-Jan-2011 01:07:11</td>
     *         <td>"+15min"</td>
     *         <td>1-Jan-2011 01:15:00</td>
     *     </tr>
     *     <tr>
     *         <td>1-Jan-2011 01:07:11</td>
     *         <td>"-15min"</td>
     *         <td>1-Jan-2011 01:00:00</td>
     *     </tr>
     *     <tr>
     *         <td>1-Jan-2011 01:14:11</td>
     *         <td>"-15min"</td>
     *         <td>1-Jan-2011 01:00:00</td>
     *     </tr>
     * </table>
     *
     * @param dateTime The date/time to round.
     * @param periodString The period string to round it to.
     * @return a new {@link ReadableDateTime} instance rounded to the specified value.
     */
    @API
    public static ReadableDateTime roundDate( ReadableDateTime dateTime, String periodString ) {
        MutableDateTime mDateTime = dateTime.toMutableDateTime();

        Matcher matcher = ROUND_PATTERN.matcher( periodString );
        if ( !matcher.matches() )
            throw new DatePeriodException( "Invalid period pattern: " + periodString );

        Shift shift = Shift.find( matcher.group( 1 ) );
        int amount = parseAmount( matcher.group( 2 ) );

        PeriodType periodType = PeriodType.forKey( matcher.group( 3 ), null );
        if ( periodType == null )
            throw new DatePeriodException( "Invalid period: " + matcher.group( 3 ) );

        // Get the property
        MutableDateTime.Property property = mDateTime.property( periodType.getDateTimeFieldType() );


        // Round to this property value
        switch ( shift ) {
            case NEAREST:
                property.roundHalfCeiling();
                break;
            case UP:
                property.roundCeiling();
                break;
            case DOWN:
                property.roundFloor();
                break;
        }

        // Then we round by the amount of the period type
        int baseValue = periodType.getBaseValue();

        int propertyVal = property.get() - baseValue;
        int div = propertyVal / amount;
        int mod = propertyVal % amount;

        // round down to the 'amount' value if appropriate.
        if ( ( shift == Shift.NEAREST && mod >= (amount + amount % 2) / 2 ) || ( shift == Shift.UP && mod > 0 ) )
            div++;

        property.set( div * amount + baseValue );

        return mDateTime;
    }

    private static int parseAmount( String amount ) {
        try {
            return Integer.parseInt( amount );
        } catch ( NumberFormatException e ) {
            throw new DatePeriodException( "Invalid amount value: " + amount );
        }
    }
}
