package org.randombits.utils.lang;

import java.lang.reflect.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Provides utility functions related to {@link Class}es.
 */
@API
public class ClassUtils {

    /**
     * Get the underlying class for a handler, or null if the handler is a variable
     * handler.
     *
     * @param type the handler
     * @return the underlying class
     */
    private static Class<?> getClass( Type type ) {
        if ( type instanceof Class<?> ) {
            return (Class<?>) type;
        } else if ( type instanceof ParameterizedType ) {
            return getClass( ( (ParameterizedType) type ).getRawType() );
        } else if ( type instanceof GenericArrayType ) {
            Type componentType = ( (GenericArrayType) type ).getGenericComponentType();
            Class<?> componentClass = getClass( componentType );
            if ( componentClass != null ) {
                return Array.newInstance( componentClass, 0 ).getClass();
            } else {
                return null;
            }
        } else {
            return null;
        }
    }

    /**
     * Get the actual handler arguments a child class has used to extend a generic
     * base class. Typical use in an abstract superclass class constructor would be:
     * <p/>
     * <code>
     * public abstract class Foo&lt;Bar&gt; {
     * private final Class&lt;Bar&gt; barClass;
     * public Foo() {
     * barClass = (Class&lt;Bar&gt;) ClassUtils.getTypeArguments( Foo.class, getClass() ).get( 0 );
     * }
     * }
     * </code>
     * <p/>
     * This will populate the <code>barClass</code> instance with whatever actual concrete class for 'Bar' is defined
     * in the subclass.
     *
     * @param baseClass  the base class
     * @param childClass the child class
     * @return a list of the raw classes for the actual handler arguments.
     */
    @API
    public static <T> List<Class<?>> getTypeArguments( Class<T> baseClass, Class<? extends T> childClass ) {
        Map<Type, Type> resolvedTypes = new HashMap<Type, Type>();
        Type type = childClass;
        // start walking up the inheritance hierarchy until we hit baseClass
        while ( !getClass( type ).equals( baseClass ) ) {
            if ( type instanceof Class<?> ) {
                // there is no useful information for us in raw types, so just
                // keep going.
                type = ( (Class<?>) type ).getGenericSuperclass();
            } else {
                ParameterizedType parameterizedType = (ParameterizedType) type;
                assert parameterizedType != null;
                Class<?> rawType = (Class<?>) parameterizedType.getRawType();

                Type[] actualTypeArguments = parameterizedType.getActualTypeArguments();
                TypeVariable<?>[] typeParameters = rawType.getTypeParameters();
                for ( int i = 0; i < actualTypeArguments.length; i++ ) {
                    resolvedTypes.put( typeParameters[i], actualTypeArguments[i] );
                }

                if ( !rawType.equals( baseClass ) ) {
                    type = rawType.getGenericSuperclass();
                }
            }
        }

        // finally, for each actual handler argument provided to baseClass,
        // determine (if possible)
        // the raw class for that handler argument.
        Type[] actualTypeArguments;
        if ( type instanceof Class<?> ) {
            actualTypeArguments = ( (Class<?>) type ).getTypeParameters();
        } else {
            assert type != null;
            actualTypeArguments = ( (ParameterizedType) type ).getActualTypeArguments();
        }
        List<Class<?>> typeArgumentsAsClasses = new ArrayList<Class<?>>();
        // resolve types by chasing down handler variables.
        for ( Type baseType : actualTypeArguments ) {
            while ( resolvedTypes.containsKey( baseType ) ) {
                baseType = resolvedTypes.get( baseType );
            }
            typeArgumentsAsClasses.add( getClass( baseType ) );
        }
        return typeArgumentsAsClasses;
    }

}
