/* Copyright (c) 2002, David Peterson
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *  -  Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer. 
 *
 *  -  Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution. 
 *
 *  -  Neither the name of randombits.org nor the names of its contributors may
 *     be used to endorse or promote products derived from this software without
 *     specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package org.randombits.utils.text;

import org.randombits.utils.lang.API;

import java.io.Serializable;

/**
 * TextBuffer is a replacement (of sorts) for the String and StringBuffer
 * classes. The main differences are as follows: <p/>
 * <ul>
 * <li>Non-Final - This class can be overridden to modify or enhance
 * functionality.</li>
 * <li>Mutable - The contents of this class can be modified, ala StringBuffer.</li>
 * <li>Larger feature set - This class has several extra functions that neither
 * class has.</li>
 * <li>Unsynchronized - Unlike StringBuffer, this class is not thread-safe.
 * However it does not have the overhead of synchronization.</li>
 * </ul>
 * <p/> Some basic design policies:
 * <ul>
 * <li>All methods that change the string value will return the same
 * TextBuffer. Eg. {@link #add(char)}, {@link #remove(int)}.</li>
 * <li>There will be equivalent methods on TextBuffer for all actions available
 * for String</li>
 * </ul>
 *
 * @author David Peterson
 */
@API
public class TextBuffer implements CharSequence, Cloneable, Comparable<Object>, Serializable {

    protected static final int UNDEFINED = -1;

    // The index of the alphabet represented by characters.
    private static final int ASIZE = (int) Character.MAX_VALUE;

    private int charShift[];

    protected char[] chars;

    protected int length;

    protected int increment = UNDEFINED;

    /**
     * Creates new TextBuffer
     */
    @API
    public TextBuffer() {
        this( (char[]) null );
    }

    @API
    public TextBuffer( CharSequence value ) {
        this( value.toString().toCharArray() );
    }

    @API
    public TextBuffer( char[] value ) {
        this( value, 0, ( value != null ) ? value.length : 0 );
    }

    /**
     * @param value  The array of characters to base the buffer on.
     * @param length The length of the text buffer. May be less than or equal to the length of the value array.
     * @pre length >= 0 && length <= value.length
     */
    @API
    public TextBuffer( char[] value, int length ) {
        this( value, 0, length );
    }

    /**
     * @param value  The array of characters to base the buffer on.
     * @param start  The starting index.
     * @param length The length of the number of characters to add.
     * @pre index >= 0 && index <= value.length
     * @pre start >= 0 && start < value.length
     */
    @API
    public TextBuffer( char[] value, int start, int length ) {
        chars = new char[length];
        if ( value != null ) {
            System.arraycopy( value, start, chars, 0, length );
        }
        this.length = length;
    }

    /**
     * Creates a clone of the provided TextBuffer.
     *
     * @param value The text buffer to copy
     * @pre value != null
     */
    @API
    public TextBuffer( TextBuffer value ) {
        chars = new char[value.chars.length];
        System.arraycopy( value.chars, 0, chars, 0, value.length );
        length = value.length;
        increment = value.increment;
    }

    @API
    public int indexOf( CharSequence substring, int beginIndex ) {
        return indexOf( substring.toString().toCharArray(), beginIndex );
    }

    @API
    public int indexOf( CharSequence substring ) {
        return indexOf( substring.toString(), 0 );
    }

    @API
    public int indexOf( char chr, int beginIndex ) {
        int foundAt = beginIndex;
        while ( chars[foundAt] != chr && foundAt < length )
            foundAt++;
        return ( foundAt < length ) ? foundAt : UNDEFINED;
    }

    @API
    public int indexOf( char chr ) {
        return indexOf( chr, 0 );
    }

    @API
    public int indexOf( TextBuffer substring, int beginIndex ) {
        return indexOf( substring.chars, beginIndex );
    }

    @API
    public int indexOf( TextBuffer substring ) {
        return indexOf( substring, 0 );
    }

    @API
    public int indexOf( char[] target, int beginIndex ) {

        int i, cs, stop;
        int ts = target.length;
        boolean matched = false;

        if ( charShift == null )
            charShift = new int[ASIZE];

        /* Preprocessing */
        for ( i = 0; i < ts; i++ )
            charShift[(int) target[i]] = ts - i;

        /* Searching */
        i = beginIndex;
        stop = length - ts;
        while ( !matched && i <= stop ) {
            if ( match( chars, i, target, 0, ts ) ) {
                matched = true;
            } else if ( i < stop ) {
                cs = charShift[(int) chars[i + ts]];
                i += ( cs != 0 ) ? cs : ts + 1; /* shift */
            } else {
                i += ts + 1;
            }
        }

        for ( int j = 0; j < ts; j++ )
            charShift[(int) target[j]] = 0;

        return ( matched ) ? i : -1;
    }

    @API
    public int indexOf( char[] target ) {
        return indexOf( target, 0 );
    }

    /**
     * Returns a <b>new</b> TextBuffer that contains the characters between
     * <code>beginIndex</code> (inclusive) and <code>endIndex</code>
     * (exclusive).
     * <p/>
     * <b>Note:</b> Unlike most other methods in TextBuffer, this one returns a
     * new instance of TextBuffer, rather than modifying the current one.
     *
     * @param beginIndex the starting index (inclusive)
     * @param endIndex   the end index (exclusive)
     * @return A <b>new</b> TextBuffer with the specified substring.
     * @throws IndexOutOfBoundsException if the <code>beginIndex</code> is negative, or
     *                                   <code>endIndex</code> is larger than the length of this
     *                                   TextBuffer object, or <code>beginIndex</code> is larger
     *                                   than <code>endIndex</code>
     */
    @API
    public TextBuffer substring( int beginIndex, int endIndex ) {
        if ( beginIndex < 0 )
            throw new IndexOutOfBoundsException( beginIndex + " < 0" );
        else if ( endIndex > length )
            throw new IndexOutOfBoundsException( endIndex + " > length" );
        else if ( beginIndex > endIndex )
            throw new IndexOutOfBoundsException( beginIndex + " > " + endIndex );

        return new TextBuffer( chars, beginIndex, endIndex - beginIndex );
    }

    /**
     * Returns a <b>new</b> TextBuffer that contains the characters from
     * <code>beginIndex</code> (inclusive) to the end of the buffer.
     * <p/>
     * <b>Note:</b> Unlike most other methods in TextBuffer, this one returns a
     * new instance of TextBuffer, rather than modifying the current one.
     *
     * @param beginIndex the starting index (inclusive)
     * @return A <b>new</b> TextBuffer with the specified substring.
     * @throws IndexOutOfBoundsException if the <code>beginIndex</code> is negative or is larger
     *                                   than the length of this TextBuffer object
     */
    @API
    public TextBuffer substring( int beginIndex ) {
        return substring( beginIndex, length );
    }

    /**
     * Adds the substring into this string, at the specified index.
     *
     * @param index     the index to insert at
     * @param substring the substring to insert
     * @return this TextBuffer
     * @pre index >= 0 && index <= length
     * @pre substring != null
     * @post indexOf(substring, index) == index
     */
    @API
    public TextBuffer add( int index, CharSequence substring ) {
        return add( index, substring.toString().toCharArray() );
    }

    /**
     * Adds the substring to the end of this string.
     *
     * @param substring the substring to insert
     * @return this TextBuffer
     */
    @API
    public TextBuffer add( CharSequence substring ) {
        return add( length, substring );
    }

    /**
     * Adds the substring into this string, at the specified index.
     *
     * @param index     the index to insert at
     * @param substring the substring to insert
     * @return this TextBuffer
     */
    @API
    public TextBuffer add( int index, TextBuffer substring ) {
        return add( index, substring.chars, 0, substring.length );
    }

    /**
     * Adds the substring to the end of this string.
     *
     * @param substring the substring to insert
     * @return this TextBuffer
     */
    @API
    public TextBuffer add( TextBuffer substring ) {
        return add( length, substring );
    }

    /**
     * Adds the substring into this string, at the specified index.
     *
     * @param index     the index to insert at
     * @param substring the substring to insert
     * @return this TextBuffer
     * @pre index >= 0 && index <= length
     * @pre substring != null
     * @post indexOf(substring, index) == index
     */
    @API
    public TextBuffer add( int index, char[] substring ) {
        return add( index, substring, 0, substring.length );
    }

    /**
     * Adds the substring to the end of this string.
     *
     * @param substring the substring to insert
     * @return this TextBuffer
     * @pre substring != null
     */
    @API
    public TextBuffer add( char[] substring ) {
        return add( length, substring );
    }

    /**
     * Adds the substring into this string, at the specified index.
     *
     * @param index     the index to insert at
     * @param substring the substring to add
     * @param startAt   the index to start adding from substring
     * @param count     the number of characters after startAt to add from substring
     * @return this TextBuffer
     * @pre index >= 0 && index <= length
     * @pre substring != null
     * @pre startAt >=0 && startAt < substring.length
     * @pre count >= 0 && startAt+count <= substring.length
     * @post indexOf(substring, index) == index
     */
    @API
    public TextBuffer add( int index, char[] substring, int startAt, int count ) {
        // Ensure the chars array is big enough.
        ensureCharsLength( length + count );
        // move the characters currently in position.
        if ( index < length )
            moveChars( index, index + count, length - index );
        System.arraycopy( substring, startAt, chars, index, count );
        length += count;

        return this;
    }

    @API
    public TextBuffer add( char ch ) {
        return add( length, new char[]{ch}, 0, 1 );
    }

    @API
    public TextBuffer add( int index, char ch ) {
        return add( index, new char[]{ch}, 0, 1 );
    }

    /**
     * Removes the <code>count</code> characters from this string, starting at
     * <code>fromIndex</code>.
     *
     * @param fromIndex the index to start removing from
     * @param count     the number of characters to remove from the string
     * @return this TextBuffer, with the specified segment removed
     * @pre fromIndex >= 0 && fromIndex < length
     * @pre count >= 0
     */
    @API
    public TextBuffer remove( int fromIndex, int count ) {
        // make sure count doesn't overstep the bounds
        int toIndex = fromIndex + count;
        count = ( toIndex >= length ) ? length - fromIndex : count;
        if ( fromIndex < length )
            moveChars( toIndex, fromIndex, length - toIndex );
        length -= count;

        return this;
    }

    /**
     * Removes the remaining characters from this string, starting at
     * <code>fromIndex</code>.
     *
     * @param fromIndex the index to start removing from
     * @return this TextBuffer, with the specified segment removed
     * @pre fromIndex >= 0 && fromIndex < length
     */
    @API
    public TextBuffer remove( int fromIndex ) {
        return remove( fromIndex, length - fromIndex );
    }

    @API
    public TextBuffer replace( char oldChar, char newChar ) {
        return replace( oldChar, newChar, -1 );
    }

    @API
    public TextBuffer replace( char oldChar, char newChar, int instances ) {
        for ( int i = 0; i < length && instances != 0; i++ ) {
            if ( chars[i] == oldChar ) {
                chars[i] = newChar;
                instances -= ( instances > 0 ) ? 1 : 0;
            }
        }
        return this;
    }

    /**
     * Replaces up to <code>instances</code> occurrences of
     * <code>oldSubstr</code> with <code>newSubstr</code>, from left to
     * right. To replace all occurrences, pass in -1 as the value of
     * <code>instances</code>.
     *
     * @param oldSubstr the substring to replace
     * @param newSubstr the substring to put in its place
     * @param instances the number of times to replace any occurrences of the string
     * @return The text buffer with the old value replaced with the new one.
     * @pre instances >= -1
     * @pre oldSubstr != null
     * @pre newSubstr != null
     */
    @API
    public TextBuffer replace( CharSequence oldSubstr, CharSequence newSubstr, int instances ) {
        return replace( oldSubstr.toString().toCharArray(), newSubstr.toString().toCharArray(), instances );
    }

    /**
     * Replaces all occurrences of <code>oldSubstr</code> with
     * <code>newSubstr</code>, from left to right.
     *
     * @param oldSubstr the substring to replace
     * @param newSubstr the substring to put in its place
     * @return the text buffer with the old substring replaced with the new.
     * @pre oldSubstr != null
     * @pre newSubstr != null
     */
    @API
    public TextBuffer replace( CharSequence oldSubstr, CharSequence newSubstr ) {
        return replace( oldSubstr, newSubstr, -1 );
    }

    /**
     * Replaces up to <code>instances</code> occurrences of
     * <code>oldSubstr</code> with <code>newSubstr</code>, from left to
     * right. To replace all occurrences, pass in -1 as the value of
     * <code>instances</code>.
     *
     * @param oldSubstr the substring to replace
     * @param newSubstr the substring to put in its place
     * @param instances the number of times to replace any occurrences of the string
     * @return the text buffer with the old substring replaced with new.
     * @pre instances >= -1
     * @pre oldSubstr != null
     * @pre newSubstr != null
     */
    @API
    public TextBuffer replace( TextBuffer oldSubstr, TextBuffer newSubstr, int instances ) {
        return replace( oldSubstr.toCharArray(), newSubstr.toCharArray(), instances );
    }

    /**
     * Replaces all occurrences of <code>oldSubstr</code> with
     * <code>newSubstr</code>, from left to right.
     *
     * @param oldSubstr the substring to replace
     * @param newSubstr the substring to put in its place
     * @return the text buffer with the old substring replaced by the new.
     * @pre oldSubstr != null
     * @pre newSubstr != null
     */
    @API
    public TextBuffer replace( TextBuffer oldSubstr, TextBuffer newSubstr ) {
        return replace( oldSubstr, newSubstr, -1 );
    }

    /**
     * Replaces up to <code>instances</code> occurrences of
     * <code>oldSubstr</code> with <code>newSubstr</code>, from left to
     * right. To replace all occurrences, pass in -1 as the value of
     * <code>instances</code>.
     *
     * @param oldSubstr the substring to replace
     * @param newSubstr the substring to put in its place
     * @param instances the number of times to replace any occurrences of the string
     * @return the text buffer, with the old stubstring replaced with the new.
     * @pre instances >= -1
     * @pre oldSubstr != null
     * @pre newSubstr != null
     */
    @API
    public TextBuffer replace( char[] oldSubstr, char[] newSubstr, int instances ) {
        int nextIndex = indexOf( oldSubstr );
        int diff = newSubstr.length - oldSubstr.length;
        int oldLength = oldSubstr.length, newLength = newSubstr.length;
        while ( nextIndex != -1 && ( instances == -1 || instances > 0 ) ) {
            ensureCharsLength( length + diff );
            // move the rest of the string
            moveChars( nextIndex + oldLength, nextIndex + newLength, length - ( nextIndex + oldLength ) );
            System.arraycopy( newSubstr, 0, chars, nextIndex, newLength );
            length += diff;
            nextIndex = indexOf( oldSubstr, nextIndex + newLength );
            instances -= ( instances > 0 ) ? 1 : 0;
        }
        return this;
    }

    /**
     * Replaces all occurrences of <code>oldSubstr</code> with
     * <code>newSubstr</code>, from left to right.
     *
     * @param oldSubstr the substring to replace
     * @param newSubstr the substring to put in its place
     * @return the text buffer, with the old stubstring replaced with the new.
     * @pre oldSubstr != null
     * @pre newSubstr != null
     */
    @API
    public TextBuffer replace( char[] oldSubstr, char[] newSubstr ) {
        return replace( oldSubstr, newSubstr, -1 );
    }

    /**
     * Pads the string so that it is at least <code>minSize</code> characters
     * long. Any extra characters needed are set to <code>val</code>. They
     * are always prefixed.
     *
     * @param minSize the minimum length the string may be after this call
     * @param val     the character to pad with
     * @return this string
     */
    @API
    public TextBuffer pad( int minSize, char val ) {
        return pad( minSize, val, true );
    }

    /**
     * Pads the string so that it is at least <code>minSize</code> characters
     * long. Any extra characters needed are set to <code>val</code>. If
     * <code>prefix</code> is true, the characters are added to the front of
     * the string, if false, they are added to the end of the string.
     *
     * @param minSize the minimum length the string may be after this call
     * @param val     the character to pad with
     * @param prefix  if true, prefix, otherwise postfix
     * @return this string
     */
    @API
    public TextBuffer pad( int minSize, char val, boolean prefix ) {
        if ( minSize > length ) {
            ensureCharsLength( minSize );
            int index, padSize = minSize - length;
            if ( prefix ) {
                index = 0;
                moveChars( 0, padSize, length );
            } else {
                index = length;
            }

            for ( int i = index; i < index + padSize; i++ )
                chars[i] = val;
        }
        return this;
    }

    /**
     * Trims any white-space from both ends of the string. White-space includes
     * spaces (' '), tabs ('\t'), and new-lines ('\r', '\n').
     *
     * @return this string, trimmed.
     */
    @API
    public TextBuffer trim() {
        boolean stopped = false;
        int fromIndex = 0, toIndex = length - 1;
        while ( !stopped ) {
            switch ( chars[fromIndex] ) {
                case ' ':
                case '\t':
                case '\n':
                case '\r':
                    fromIndex++;
                    break;
                default:
                    stopped = true;
                    break;
            }
        }

        stopped = false;
        while ( !stopped ) {
            switch ( chars[toIndex] ) {
                case ' ':
                case '\t':
                case '\n':
                case '\r':
                    toIndex--;
                    break;
                default:
                    stopped = true;
                    break;
            }
        }

        length = toIndex - fromIndex;
        if ( fromIndex != 0 )
            moveChars( fromIndex, 0, length );

        return this;
    }

    /**
     * Converts the string to lower case.
     *
     * @return this TextBuffer object, converted to lower case.
     */
    @API
    public TextBuffer toLowerCase() {
        for ( int i = 0; i < length; i++ )
            chars[i] = Character.toLowerCase( chars[i] );
        return this;
    }

    /**
     * Converts the string to upper case.
     *
     * @return this TextBuffer object, converted to upper case.
     */
    @API
    public TextBuffer toUpperCase() {
        for ( int i = 0; i < length; i++ )
            chars[i] = Character.toUpperCase( chars[i] );
        return this;
    }

    /**
     * Converts the string to upper case.
     *
     * @return this TextBuffer object, converted to upper case.
     */
    @API
    public TextBuffer toTitleCase() {
        for ( int i = 0; i < length; i++ )
            chars[i] = Character.toTitleCase( chars[i] );
        return this;
    }

    /**
     * Returns the English plural of the current string. The plural is based on
     * the last character(s) of the string. If it ends with an 's' or 'sh', the
     * plural will end with 'es'. If it ends with a 'y', the plural will end
     * with 'ies'. All other cases end with 's'.
     *
     * @return The English plural of the last word in the string.
     */
    @API
    public TextBuffer plural() {
        String extra = "s";
        if ( this.endsWith( "sh" ) || this.endsWith( "s" ) || this.endsWith( "ch" ) || this.endsWith( "x" ) ) {
            extra = "es";
        } else if ( this.endsWith( "y" ) ) {
            if ( length > 1 && "aeiou".indexOf( Character.toLowerCase( chars[length - 2] ) ) == -1 ) {
                remove( length - 1 );
                extra = "ies";
            }
        }
        return add( extra );
    }

    @API
    public boolean startsWith( CharSequence substring ) {
        return startsWith( substring.toString().toCharArray() );
    }

    @API
    public boolean startsWith( TextBuffer substring ) {
        return startsWith( substring.chars, 0, substring.length );
    }

    @API
    public boolean startsWith( char[] substring ) {
        return startsWith( substring, 0, substring.length );
    }

    @API
    public boolean startsWith( char[] substring, int offset, int count ) {
        return count <= length && match( chars, 0, substring, offset, count );
    }

    @API
    public boolean endsWith( CharSequence substring ) {
        return endsWith( substring.toString().toCharArray() );
    }

    @API
    public boolean endsWith( TextBuffer substring ) {
        return endsWith( substring.chars, 0, substring.length );
    }

    @API
    public boolean endsWith( char[] substring ) {
        return endsWith( substring, 0, substring.length );
    }

    @API
    public boolean endsWith( char[] substring, int offset, int count ) {
        return count <= length && match( chars, length - count, substring, offset, count );
    }

    @API
    public byte[] getBytes() {
        return toString().getBytes();
    }

    @API
    public byte[] getBytes( String enc ) throws java.io.UnsupportedEncodingException {
        return toString().getBytes( enc );
    }

    @API
    public void getChars( int srcBegin, int srcEnd, char[] dst, int dstBegin ) {
        if ( srcEnd < srcBegin )
            throw new IndexOutOfBoundsException( "srcEnd (" + srcEnd + ") is less than srcBegin (" + srcBegin
                    + ")" );
        if ( srcEnd > length )
            throw new IndexOutOfBoundsException( "srcEnd (" + srcEnd + ") exceeds the length of the TextBuffer ("
                    + length + ")" );
        System.arraycopy( chars, srcBegin, dst, dstBegin, srcEnd - srcBegin );
    }

    @Override
    @API
    public char charAt( int index ) {
        if ( index > length )
            throw new IndexOutOfBoundsException( "index (" + index
                    + ") is greater than the length of the TextBuffer (" + length + ")" );
        return chars[index];
    }

    @Override
    @API
    public CharSequence subSequence( int start, int end ) {
        return substring( start, end );
    }

    /**
     * Compares the specified object to this TextBuffer.
     *
     * @param obj the object to compare to.
     * @return a negative value if <code>obj</code> is less than this string,
     *         0 if they are equal or a positive value if <code>obj</code> is
     *         greater.
     * @throws NullPointerException if <code>obj</code> is null.
     * @throws ClassCastException   if <code>obj</code> is not an instance of String or
     *                              TextBuffer.
     * @see #compareTo(int, int, char[], int, int, boolean)
     */

    @Override
    @API
    public int compareTo( Object obj ) {
        if ( obj instanceof String ) {
            return compareTo( (String) obj );
        } else if ( obj instanceof TextBuffer ) {
            return compareTo( (TextBuffer) obj );
        } else if ( obj instanceof CharSequence ) {
            return compareTo( obj.toString() );
        }
        throw new ClassCastException(
                "Only Strings, CharSequences and CharBuffers may be compared to CharBuffers." );
    }

    /**
     * Compares the specified String to this TextBuffer.
     *
     * @param val the String to compare to.
     * @return a negative value if <code>val</code> is less than this string,
     *         0 if they are equal or a positive value if <code>val</code> is
     *         greater.
     * @throws NullPointerException if <code>val</code> is null.
     * @see #compareTo(int, int, char[], int, int, boolean)
     */
    @API
    public int compareTo( String val ) {
        char strChars[] = val.toCharArray();
        return compareTo( 0, length, strChars, 0, strChars.length, false );
    }

    /**
     * Compares the specified TextBuffer to this TextBuffer.
     *
     * @param val the TextBuffer to compare to.
     * @return a negative value if <code>val</code> is less than this string,
     *         0 if they are equal or a positive value if <code>val</code> is
     *         greater.
     * @throws NullPointerException if <code>val</code> is null.
     * @see #compareTo(int, int, char[], int, int, boolean)
     */
    @API
    public int compareTo( TextBuffer val ) {
        return compareTo( 0, length, val.chars, 0, val.length, false );
    }

    /**
     * Compares the specified TextBuffer to this TextBuffer, ignoring any case
     * differences.
     *
     * @param val the TextBuffer to compare to.
     * @return a negative value if <code>val</code> is less than this string,
     *         0 if they are equal or a positive value if <code>val</code> is
     *         greater.
     * @throws NullPointerException if <code>val</code> is null.
     * @see #compareTo(int, int, char[], int, int, boolean)
     */
    @API
    public int compareToIgnoreCase( TextBuffer val ) {
        return compareTo( 0, length, val.chars, 0, val.length, true );
    }

    /**
     * Compares the specified character array to this TextBuffer.
     *
     * @param val the character array to compare to.
     * @return a negative value if <code>val</code> is less than this string,
     *         0 if they are equal or a positive value if <code> val is greater.
     * @throws NullPointerException if <code>val</code> is null.
     * @see #compareTo(int, int, char[], int, int, boolean)
     */
    @API
    public int compareTo( char[] val ) {
        return compareTo( 0, length, val, 0, val.length, false );
    }

    /**
     * Compares the specified <code>val</code> to this TextBuffer. This is the
     * mother of all compareTo methods.
     *
     * @param offset     the offset to start comparing this string from
     * @param count      the size of the substring of this string to compare to
     * @param val        the value to compare to
     * @param voffset    the offset to start comparing the value from
     * @param vcount     the size of the substring of the value to compare to
     * @param ignoreCase if true, ignore the case of both strings.
     * @return a negative value if <code>val</code> is less than this string,
     *         0 if they are equal or a positive value if <code> val is greater.
     * @throws IndexOutOfBoundsException if any of the following is true:
     *                                   <ul>
     *                                   <li>offset is negative</li>
     *                                   <li>offset is greater than the length of this string</li>
     *                                   <li>offset+count is greater than the length of this string</li>
     *                                   <li>voffset is negative</li>
     *                                   <li>voffset is greater than the length of this string</li>
     *                                   <li>voffset+vcount is larger than val.length</li>
     *                                   </ul>
     * @throws NullPointerException      if <code>val</code> is null.
     */
    @API
    public int compareTo( int offset, int count, char[] val, int voffset, int vcount, boolean ignoreCase ) {
        int max = offset + count, vmax = voffset + vcount;
        if ( max > length )
            throw new IndexOutOfBoundsException( "offset + count (" + offset + count
                    + ") exceeds the length of this TextBuffer (" + length + ")" );
        if ( vmax > val.length )
            throw new IndexOutOfBoundsException( "voffset + vcount (" + voffset + vcount
                    + ") exceeds the length of the val (" + val.length + ")" );
        while ( offset < max && voffset < vmax
                && uncase( chars[offset], ignoreCase ) == uncase( val[voffset], ignoreCase ) ) {
            offset++;
            voffset++;
        }

        if ( offset == max && voffset == vmax ) { // both
            return 0;
        } else if ( offset == max || voffset == vmax ) { // one or the other
            return max - vmax;
        } else { // character mismatch
            return uncase( chars[offset], ignoreCase ) - uncase( val[voffset], ignoreCase );
        }
    }

    /**
     * Returns true if the object is an TextBuffer and it contains the exact
     * same characters as this string. Other 'environmental' conditions are not
     * compared.
     */
    @Override
    @API
    public boolean equals( Object obj ) {
        return obj instanceof TextBuffer && compareTo( (TextBuffer) obj ) == 0;
    }

    /**
     * Returns true if the object is an TextBuffer and it contains the same
     * characters (ignoring case) as this string. Other 'environmental'
     * conditions are not compared.
     *
     * @param obj The object to compare it to.
     * @return true if the object is an instance of TextBuffer and the values are equal, ignoring case.
     */
    @API
    public boolean equalsIgnoreCase( Object obj ) {
        if ( obj instanceof TextBuffer ) {
            return compareToIgnoreCase( (TextBuffer) obj ) == 0;
        }
        return false;
    }

    @Override
    @API
    public int length() {
        return length;
    }

    @API
    @Override
    public String toString() {
        return new String( chars, 0, length );
    }

    @API
    public char[] toCharArray() {
        char[] ret = new char[length];
        System.arraycopy( chars, 0, ret, 0, length );
        return ret;
    }

    // ///// Public static methods ////////

    @API
    public static TextBuffer valueOf( boolean val ) {
        return new TextBuffer( val ? "true" : "false" );
    }

    @API
    public static TextBuffer valueOf( int val ) {
        return new TextBuffer( String.valueOf( val ) );
    }

    @API
    public static TextBuffer valueOf( long val ) {
        return new TextBuffer( String.valueOf( val ) );
    }

    @API
    public static TextBuffer valueOf( float val ) {
        return new TextBuffer( String.valueOf( val ) );
    }

    @API
    public static TextBuffer valueOf( double val ) {
        return new TextBuffer( String.valueOf( val ) );
    }

    @API
    public static TextBuffer valueOf( char val ) {
        return new TextBuffer( new char[]{val} );
    }

    @API
    public static TextBuffer valueOf( char[] val ) {
        return new TextBuffer( val );
    }

    @API
    public static TextBuffer valueOf( char[] val, int offset, int count ) {
        return new TextBuffer( val, offset, count );
    }

    @API
    public static TextBuffer valueOf( Object val ) {
        return new TextBuffer( val.toString() );
    }

    // ///// Private functions ////////

    /**
     * Returns an un-case-sensitive version of the character, if appropriate.
     *
     * @param c          the character to uncase.
     * @param ignoreCase if set to true, don't modify the character.
     * @return The uncased character.
     */
    @API
    private static char uncase( char c, boolean ignoreCase ) {
        return ( ignoreCase ) ? Character.toUpperCase( Character.toLowerCase( c ) ) : c;
    }

    /**
     * @param a      The first character array.
     * @param as     The index to start examining the first character array.
     * @param b      The second character array.
     * @param bs     The index to start examining the second character array.
     * @param length The length to compare across.
     * @return <code>true</code> if both sets of characters are equal.
     * @pre a != null
     * @pre b != null
     */
    @API
    private static boolean match( char[] a, int as, char[] b, int bs, int length ) {
        int amax = as + length;
        int bmax = bs + length;
        if ( a.length >= amax && b.length >= bmax ) {
            boolean match = true;
            while ( match && as < amax )
                match = a[as++] == b[bs++];

            return match;
        }
        return false;
    }

    /**
     * Moves characters from one point in the buffer to another point.
     *
     * @param from  The original location of the characters.
     * @param to    The target location of the characters.
     * @param count The number of characters to move.
     * @pre from >=0 && from + count < length
     * @pre to >= 0 && to + count < length
     */
    @API
    private void moveChars( int from, int to, int count ) {
        if ( from != to ) {
            boolean forward = from < to;
            int inc = ( forward ) ? -1 : 1;
            int start = ( forward ) ? count - 1 : 0;
            int finish = ( forward ) ? -1 : count;

            for ( int i = start; i != finish; i += inc )
                chars[to + i] = chars[from + i];
        }
    }

    @API
    private void ensureCharsLength( int required ) {
        if ( chars.length < required ) {
            int newLength = ( chars.length == 0 ) ? 1 : chars.length;
            while ( newLength < required )
                newLength += ( increment > 0 ) ? increment : newLength;

            char newChars[] = new char[newLength];
            System.arraycopy( chars, 0, newChars, 0, chars.length );
            chars = newChars;
        }
    }

    /**
     * Returns a copy of this TextBuffer.
     *
     * @return returns a new TextBuffer that is a copy of this text buffer.
     */
    @API
    public TextBuffer copy() {
        return new TextBuffer( this );
    }

    /**
     * Returns a copy of this TextBuffer.
     */
    @Override
    @API
    public Object clone() throws CloneNotSupportedException {
        return copy();
    }
}
