package org.randombits.utils.text;

import org.randombits.utils.lang.API;

import java.nio.CharBuffer;
import java.text.Collator;
import java.util.Comparator;
import java.util.Locale;

/**
 * This is an implementation of {@link java.util.Comparator} which allows for flexible
 * rules about comparing String values. In particular, there are several
 * preconfigured comparators available for performing locale-sensitive
 * 'natural' order sorting, where sequences of numbers are ordered according
 * to their numerical order, rather than their character order. For example,
 * if sorting by character order, '3' comes after '21'. When sorted by numerical,
 * or 'natural' order, '3' comes before '21'.
 *
 * @author David Peterson
 * @see #getNaturalInstance()
 * @see #getNaturalInstance(java.util.Locale)
 * @see #getNaturalInstance(java.text.Collator)
 */
@API
public class TextComparator implements Comparator<Object> {

    /**
     * Creates a {@link java.util.Comparator} which will sort Strings into the following
     * order:
     * <p/>
     * <ol>
     * <li><b>symbols</b> - Non-alphanumeric characters. Sorted by bit order.</li>
     * <li><b>digits</b> - Sequences of positive integers. Sorted as whole
     * numbers.</li>
     * <li><b>letters</b> - Sequences of letters, as defined by
     * {@link Character#isLetter(char)}. Sorted by bit order.</li>
     * </ol>
     *
     * @return a 'natural' sort instance for the default locale.
     */
    @API
    public static TextComparator getNaturalInstance() {
        return getNaturalInstance( (Collator) null );
    }

    /**
     * Creates a {@link java.util.Comparator} which will sort Strings into the following
     * order:
     * <p/>
     * <ol>
     * <li><b>symbols</b> - Non-alphanumeric characters. Sorted by bit order.</li>
     * <li><b>digits</b> - Sequences of positive integers. Sorted as whole
     * numbers.</li>
     * <li><b>letters</b> - Sequences of letters, as defined by
     * {@link Character#isLetter(char)}. Sorted by the standard collator for
     * the locale, or by bit order if none is provided.</li>
     * </ol>
     *
     * @param locale The locale to use when comparing.
     * @return The local-specific text comparator.
     */
    @API
    public static TextComparator getNaturalInstance( Locale locale ) {
        return getNaturalInstance( locale == null ? null : Collator.getInstance( locale ) );
    }

    /**
     * Creates a {@link java.util.Comparator} which will sort Strings into the following
     * order:
     * <p/>
     * <ol>
     * <li><b>symbols</b> - Non-alphanumeric characters. Sorted by bit order.</li>
     * <li><b>digits</b> - Sequences of positive integers. Sorted as whole
     * numbers.</li>
     * <li><b>letters</b> - Sequences of letters, as defined by
     * {@link Character#isLetter(char)}. Sorted by the collator, or by bit
     * order if none is provided.</li>
     * </ol>
     *
     * @param collator The character collator.
     * @return The comparator based on the collator.
     */
    @API
    public static TextComparator getNaturalInstance( Collator collator ) {
        return new TextComparator( new TextProcessor[]{new SymbolProcessor(), new DigitProcessor(),
                new LetterProcessor( collator )} );
    }

    private TextProcessor[] processors;

    /**
     * Constructs a new TextComparator with the specified list of processors.
     * Each processor is accessed in order to determine if two Strings are
     * equal. See {@link #compare(Object, Object)} for details.
     *
     * @param processors The processor list. No items may be <code>null</code>.
     */
    @API
    public TextComparator( TextProcessor[] processors ) {
        this.processors = processors;
    }

    /**
     * Compares the two strings. This method will loop through the characters in
     * each string, checking with each processor in turn whether it is
     * supported. If both strings are supported by a single processor, the
     * processor's 'compare' method is called and if the result is not
     * <code>0</code>, the result is returned. If it is <code>0</code>, it
     * keeps progressing through the string values. If only one string can be
     * handled by the processor, it is assumed to be the winner and the
     * appropriate comparison value is returned.
     *
     * @param s1 The first string value.
     * @param s2 The second string value.
     * @return the comparison value.
     */
    @API
    public int compare( String s1, String s2 ) {
        CharBuffer cb1 = CharBuffer.wrap( s1 == null ? "" : s1 ).asReadOnlyBuffer();
        CharBuffer cb2 = CharBuffer.wrap( s2 == null ? "" : s2 ).asReadOnlyBuffer();

        TextProcessor processor;

        while ( cb1.hasRemaining() && cb2.hasRemaining() ) {
            boolean p1 = false, p2 = false;
            for ( int i = 0; !p1 && !p2 && i < processors.length; i++ ) {
                processor = processors[i];
                // If either chunk matches the current processor we may have a
                // result
                p1 = processor.canProcess( cb1.duplicate() );
                p2 = processor.canProcess( cb2.duplicate() );
                int result = 0;
                if ( p1 && p2 ) {
                    result = processor.compare( cb1, cb2 );
                    if ( result != 0 )
                        return result;
                } else if ( p1 ) {
                    return -1;
                } else if ( p2 ) {
                    return 1;
                }
            }
            // Check that at least one processor matched, otherwise return 0.
            if ( !p1 && !p2 )
                return 0;
        }

        if ( cb1.hasRemaining() )
            return 1;
        else if ( cb2.hasRemaining() )
            return -1;
        else
            return 0;
    }

    @API
    @Override
    public int compare( Object o1, Object o2 ) {
        String s1 = asString( o1 );
        String s2 = asString( o2 );
        return compare( s1, s2 );
    }

    /**
     * This method is called to convert each object to a {@link String}. By default,
     * it simply calls the 'toString' method for any non-String object. Override
     * this to provide custom conversions for specific class types.
     *
     * @param object The object.
     * @return The String, or <code>null</code> if <code>object</code> is <code>null</code>.
     */
    protected String asString( Object object ) {
        if ( object instanceof String )
            return (String) object;
        else if ( object != null )
            return object.toString();
        return null;
    }

}
