/*
 * Copyright (c) 2002, David Peterson All rights reserved. Redistribution and
 * use in source and binary forms, with or without modification, are permitted
 * provided that the following conditions are met: - Redistributions of source
 * code must retain the above copyright notice, this list of conditions and the
 * following disclaimer. - Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the distribution. -
 * Neither the name of randombits.org nor the names of its contributors may be
 * used to endorse or promote products derived from this software without
 * specific prior written permission. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT
 * HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.randombits.utils.text;

import org.randombits.utils.lang.API;
import org.randombits.utils.io.IOUtils;
import org.randombits.utils.collection.ABitSet;

import java.io.BufferedReader;
import java.io.CharArrayWriter;
import java.io.IOException;
import java.io.Reader;
import java.text.CharacterIterator;
import java.text.StringCharacterIterator;
import java.util.BitSet;

/*
 * ! <source> <file>TokenIterator.java </file> <create-date>2002-01-21 22:49
 * </create-date> <changes> </changes> <source>
 */

/**
 * @author David Peterson
 * @version 2.0
 */
@API(value = "1.0")
public class TokenIterator {

    /**
     * This class provides a method to rewind to a particular location
     * in an iterator. It is used like so:
     *
     * <pre>
     * Mark myMark = iterator.getMark();
     * // do other stuff
     * myMark.rewind();
     * </pre>
     */
    @API(value = "2.0")
    public static final class Mark {

        private final TokenIterator iterator;

        private final int index;

        private final int lastIndex;

        private Mark( TokenIterator iterator ) {
            this.iterator = iterator;
            this.index = iterator.chars.getIndex();
            this.lastIndex = iterator.lastIndex;
        }

        /**
         * Rewinds the iterator the mark was created from back to the location
         * and rewind location it was when the mark was created.
         */
        @API(value = "2.0")
        public void rewind() {
            iterator.chars.setIndex( index );
            iterator.lastIndex = lastIndex;
        }
    }

    /**
     * A BitSet with all legal characters set.
     */
    @API(value = "1.0")
    public static final BitSet ALL_CHARS = new ABitSet( Character.MAX_VALUE,
            true );

    static {
        ( (ABitSet) ALL_CHARS ).setRange( Character.MIN_VALUE, Character.MAX_VALUE );
        ( (ABitSet) ALL_CHARS ).lock();
    }

    @API(value = "1.0")
    public static final int UNBOUNDED = Integer.MAX_VALUE;

    @API(value = "1.0")
    protected final CharacterIterator chars;

    @API(value = "1.0")
    protected int lastIndex = 0;

    /**
     * Constructs a new TokenIterator.
     *
     * @param reader The reader providing the input.
     * @throws java.io.IOException if there was an I/O problem.
     */
    @API(value = "1.0")
    public TokenIterator( Reader reader ) throws IOException {
        BufferedReader in;
        if ( reader instanceof BufferedReader )
            in = (BufferedReader) reader;
        else
            in = new BufferedReader( reader );

        CharArrayWriter writer = new CharArrayWriter();
        IOUtils.pipe( in, writer );
        chars = new StringCharacterIterator( writer.toString() );
    }

    /**
     * Constructs a new TokenIterator.
     *
     * @param chars The character iterator.
     */
    @API(value = "1.0")
    public TokenIterator( CharacterIterator chars ) {
        this.chars = chars;
        this.lastIndex = chars.getBeginIndex();
    }


    @API(value = "1.0")
    public TokenIterator( String string ) {
        this( new StringCharacterIterator( string ) );
    }

    /**
     * Returns the sequence of characters that match the specified character set
     * up to but not including the <code>terminator</code>. If characters
     * that do not match the charSet are found before the terminator, or the end
     * of the string is hit, <code>null</code> is returned.
     *
     * @param charSet    The set of characters that are allowed in the returned sequence.
     * @param terminator The character that will terminate the sequence.
     * @return The sequence of characters, or <code>null</code> if no match was found.
     */
    @API(value = "1.0")
    public CharSequence getSequence( BitSet charSet, String terminator ) {
        TextBuffer buff = new TextBuffer();
        int ch;
        while ( !matchToken( terminator ) ) {
            ch = chars.current();
            if ( ch == CharacterIterator.DONE || !charSet.get( ch ) )
                return null;
            else
                buff.add( (char) ch );
            chars.next();
        }
        return buff;
    }

    /**
     * Reads a sequence of characters that match those in the <code>charSet</code>.
     * If no characters could be matched, an empty string is returned.
     *
     * @param charSet the set of legal characters that may be in this sequence
     * @return the sequence of characters read
     */
    @API(value = "1.0")
    public CharSequence getSequence( BitSet charSet ) {
        return getSequence( charSet, 0, UNBOUNDED );
    }

    @API(value = "1.0")
    public CharSequence getSequence( BitSet charSet, int min ) {
        return getSequence( charSet, min, UNBOUNDED );
    }

    /**
     * Reads a sequence of characters. If the minimum number of
     * valid characters (as defined by <code>charSet</code>) could not
     * be matched, <code>null</code> is returned. If the sequence fails,
     * the index will still be at the character which failed,
     * but it can be rolled back by calling {@link #rewind()}.
     *
     * @param charSet the set of legal characters that may be in this sequence
     * @param min     the minimum number of characters that must be read
     * @param max     the maximum number of characters to read
     * @return the sequence of characters read, or <code>null</code>
     *         if the minimum number of legal characters was not met.
     * @see #rewind()
     */
    @API(value = "1.0")
    public CharSequence getSequence( BitSet charSet, int min, int max ) {
        TextBuffer buff = new TextBuffer();
        int ch, count = 0;

        int mark = chars.getIndex();

        while ( count < max && ( ch = chars.current() ) != CharacterIterator.DONE ) {
            if ( charSet.get( ch ) ) {
                buff.add( (char) ch );
                count++;
                chars.next();
            } else {
                // reset the pointer to the previous character.
                break;
            }
        }
        if ( count < min ) {
            chars.setIndex( mark );
            return null;
        } else {
            lastIndex = mark;
            return buff;
        }
    }

    /**
     * Returns <code>true</code> if the next set of characters matches the
     * <code>token</code>, without moving the index.
     *
     * @param token The String token to check for.
     * @return <code>true</code> if the next set of characters match.
     */
    @API(value = "1.0")
    public boolean matchToken( String token ) {
        int mark = chars.getIndex();
        boolean match = getToken( token ) != null;
        chars.setIndex( mark );
        return match;
    }

    /**
     * Returns <code>true</code> if the next character in this iterator
     * matches the value of <code>ch</code>, without moving the index.
     *
     * @param ch The character to check for.
     * @return <code>true</code> if they match.
     */
    @API(value = "1.0")
    public boolean matchToken( int ch ) {
        int read = chars.current();
        return read == ch;
    }

    /**
     * Returns an equal string to <code>token</code> if the next character
     * sequence is a case-sensitive match, or <code>null</code> if it's not.
     * If the token fails to match, the position of the iterator has not moved.
     *
     * @param token The String token to check for.
     * @return An equal string, or <code>null</code>.
     */
    @API(value = "1.0")
    public CharSequence getToken( String token ) {
        return getToken( token, false );
    }

    @API(value = "1.0")
    public CharSequence getToken( String token, boolean ignoreCase ) {
        return getToken( token, ignoreCase, true );
    }

    @API(value = "2.0")
    public CharSequence getToken( String token, boolean ignoreCase, boolean rewindOnFail ) {
        int mark = chars.getIndex();

        int length = token.length();
        TextBuffer buff = new TextBuffer();
        int tok;
        for ( int i = 0; i < length; i++ ) {
            tok = getChar( token.charAt( i ), ignoreCase );
            if ( tok == CharacterIterator.DONE ) {
                if ( rewindOnFail )
                    chars.setIndex( mark ); // rewind...
                return null;
            } else {
                buff.add( (char) tok );
            }
        }

        lastIndex = mark;
        return buff;
    }

    @API(value = "1.0")
    public CharSequence getToken( int ch ) {
        return getToken( ch, false );
    }

    @API(value = "1.0")
    public CharSequence getToken( int ch, boolean ignoreCase ) {
        return getToken( ch, ignoreCase, true );
    }

    @API(value = "2.0")
    public CharSequence getToken( int ch, boolean ignoreCase, boolean rewindOnFail ) {
        int mark = chars.getIndex();
        int tok = getChar( ch, ignoreCase );

        if ( tok == CharacterIterator.DONE ) {
            if ( rewindOnFail )
                chars.setIndex( mark );
            return null;
        } else {
            lastIndex = mark;
            return String.valueOf( (char) tok );
        }
    }

    private int getChar( int ch, boolean ignoreCase ) {
        int ch_in = chars.current();
        if ( ignoreCase
                && Character.toLowerCase( (char) ch ) != Character.toLowerCase( (char) ch_in )
                || !ignoreCase && ch != ch_in ) {
            return CharacterIterator.DONE;
        }
        chars.next();
        return ch_in;
    }

    /**
     * Rewinds the iterator to the last successful operation.
     */
    @API(value = "1.0")
    public void rewind() {
        chars.setIndex( lastIndex );
    }

    /**
     * @return <code>true</code> if the TokenIterator is at the end of the set
     *         of characters being iterated.
     */
    @API(value = "1.0")
    public boolean atEnd() {
        return chars.getIndex() == chars.getEndIndex();
    }

    /**
     * @return <code>true</code> if the TokenIterator is at the beginning of
     *         the set of characters being iterated.
     */
    @API(value = "1.0")
    public boolean atBeginning() {
        return chars.getIndex() == chars.getBeginIndex();
    }

    @Override
    public String toString() {
        return viewCharsBefore().toString() + viewCharsAfter().toString();
    }

    @API(value = "2.0")
    public CharSequence getChars( int count ) {
        StringBuilder builder = new StringBuilder();

        int i = 0;
        char c;
        for ( c = chars.current(); i < count && chars.getIndex() != chars.getEndIndex(); c = chars.next() ) {
            builder.append( c );
            i++;
        }

        if ( i != count )
            return null;

        lastIndex = chars.getIndex();
        return builder;
    }

    /**
     * Gets the current index that the iterator is operating on.
     *
     * @return the current index.
     */
    @API(value = "2.0")
    public int getIndex() {
        return chars.getIndex() - chars.getBeginIndex();
    }

    /**
     * Returns the total number of characters in the sequence.
     *
     * @return The length.
     */
    @API(value = "2.0")
    public int getLength() {
        return chars.getEndIndex() - chars.getBeginIndex();
    }

    /**
     * Returns all a sequence of all characters before the current index, exclusive.
     * Does not affect the current index of the TextIterator.
     *
     * @return The before sequence.
     */
    @API(value = "2.0")
    public CharSequence viewCharsBefore() {
        int index = chars.getIndex();
        StringBuilder out = new StringBuilder();

        char c;
        for ( c = chars.first(); chars.getIndex() < index; c = chars.next() )
            out.append( c );

        return out;
    }

    /**
     * Returns all characters after (and including) the index.
     * Does not affect the current index of the TextIterator.
     *
     * @return the character sequence.
     */
    @API(value = "2.0")
    public CharSequence viewCharsAfter() {
        int index = chars.getIndex();
        StringBuilder out = new StringBuilder();

        char c;
        for ( c = chars.current(); chars.getIndex() < chars.getEndIndex(); c = chars.next() ) {
            out.append( c );
        }
        //Reset
        chars.setIndex( index );

        return out;
    }

    /**
     * Returns a {@link Mark} which will allow the iterator to be rolled
     * back to its current state, for both the current index and the {@link #rewind()}
     * location by calling {@link org.randombits.utils.text.TokenIterator.Mark#rewind()}.
     *
     * @return the new {@link Mark}.
     */
    @API(value = "2.0")
    public Mark getMark() {
        return new Mark( this );
    }
}